from dataclasses import dataclass
from typing import List

from rpe.engines import Engine
from rpe.resources import Resource


@dataclass
class Policy:
    policy_id: str
    engine: Engine
    applies_to: List[str]
    description: str = ""


@dataclass
class Evaluation:
    resource: Resource
    engine: Engine
    policy_id: str

    compliant: bool
    excluded: bool
    remediable: bool

    def remediate(self):
        return self.engine.remediate(self.resource, self.policy_id)



# Copyright 2019 The resource-policy-evaluation-library Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import collections
import pytest

from google.oauth2.credentials import Credentials

from rpe.exceptions import ResourceException
from rpe.resources.gcp import GoogleAPIResource
from rpe.resources.gcp import GcpAppEngineInstance
from rpe.resources.gcp import GcpBigqueryDataset
from rpe.resources.gcp import GcpBigtableInstance
from rpe.resources.gcp import GcpCloudFunction
from rpe.resources.gcp import GcpComputeInstance
from rpe.resources.gcp import GcpComputeDisk
from rpe.resources.gcp import GcpDatafusionInstance
from rpe.resources.gcp import GcpDataprocCluster
from rpe.resources.gcp import GcpGkeCluster
from rpe.resources.gcp import GcpGkeClusterNodepool
from rpe.resources.gcp import GcpOrganization
from rpe.resources.gcp import GcpProject
from rpe.resources.gcp import GcpProjectService
from rpe.resources.gcp import GcpPubsubSubscription
from rpe.resources.gcp import GcpPubsubTopic
from rpe.resources.gcp import GcpSqlInstance
from rpe.resources.gcp import GcpStorageBucket
from rpe.resources.gcp import GcpComputeFirewall
from rpe.resources.gcp import GcpComputeSubnetwork
from rpe.resources.gcp import GcpDataflowJob
from rpe.resources.gcp import GcpRedisInstance
from rpe.resources.gcp import GcpMemcacheInstance

test_project = "my_project"
test_resource_name = "my_resource"
client_kwargs = {
    'credentials': Credentials(token='')
}

ResourceTestCase = collections.namedtuple('ResourceTestCase', 'resource_data cls resource_type name')

test_cases = [
    ResourceTestCase(
        resource_data={
            'name': 'my_resource',
            'app': 'my_project',
            'service': 'default',
            'version': '0123456789',
            'project_id': test_project,
        },
        cls=GcpAppEngineInstance,
        resource_type='appengine.googleapis.com/Instance',
        name='//appengine.googleapis.com/apps/my_project/services/default/versions/0123456789/instances/my_resource'
    ),
    ResourceTestCase(
        resource_data={
            'name': test_resource_name,
            'project_id': test_project
        },
        cls=GcpBigqueryDataset,
        resource_type='bigquery.googleapis.com/Dataset',
        name='//bigquery.googleapis.com/projects/my_project/datasets/my_resource'
    ),
    ResourceTestCase(
        resource_data={
            'name': test_resource_name,
            'project_id': test_project
        },
        cls=GcpBigtableInstance,
        resource_type='bigtableadmin.googleapis.com/Instance',
        name='//bigtable.googleapis.com/projects/my_project/instances/my_resource'
    ),
    ResourceTestCase(
        resource_data={
            'name': test_resource_name,
            'location': 'us-central1-a',
            'project_id': test_project
        },
        cls=GcpCloudFunction,
        resource_type='cloudfunctions.googleapis.com/CloudFunction',
        name='//cloudfunctions.googleapis.com/projects/my_project/locations/us-central1-a/functions/my_resource'
    ),
    ResourceTestCase(
        resource_data={
            'name': test_resource_name,
            'location': 'us-central1-a',
            'project_id': test_project
        },
        cls=GcpComputeDisk,
        resource_type='compute.googleapis.com/Disk',
        name='//compute.googleapis.com/projects/my_project/zones/us-central1-a/disks/my_resource'
    ),
    ResourceTestCase(
        resource_data={
            'name': test_resource_name,
            'location': 'us-central1-a',
            'project_id': test_project
        },
        cls=GcpComputeInstance,
        resource_type='compute.googleapis.com/Instance',
        name='//compute.googleapis.com/projects/my_project/zones/us-central1-a/instances/my_resource'
    ),
    ResourceTestCase(
        resource_data={
            'name': test_resource_name,
            'location': 'us-central1-a',
            'project_id': test_project
        },
        cls=GcpGkeCluster,
        resource_type='container.googleapis.com/Cluster',
        name='//container.googleapis.com/projects/my_project/locations/us-central1-a/clusters/my_resource'
    ),
    ResourceTestCase(
        resource_data={
            'name': test_resource_name,
            'location': 'us-central1-a',
            'cluster': 'parent_resource',
            'project_id': test_project
        },
        cls=GcpGkeClusterNodepool,
        resource_type='container.googleapis.com/NodePool',
        name='//container.googleapis.com/projects/my_project/locations/us-central1-a/clusters/parent_resource/nodePools/my_resource'
    ),
    ResourceTestCase(
        resource_data={
            'name': '012345678901',
        },
        cls=GcpOrganization,
        resource_type='cloudresourcemanager.googleapis.com/Organization',
        name='//cloudresourcemanager.googleapis.com/organizations/012345678901'
    ),
    ResourceTestCase(
        resource_data={
            'name': test_project,
            'project_id': test_project
        },
        cls=GcpProject,
        resource_type='cloudresourcemanager.googleapis.com/Project',
        name='//cloudresourcemanager.googleapis.com/projects/my_project'
    ),
    ResourceTestCase(
        resource_data={
            'name': 'compute.googleapis.com',
            'project_id': test_project
        },
        cls=GcpProjectService,
        resource_type='serviceusage.googleapis.com/Service',
        name='//serviceusage.googleapis.com/projects/my_project/services/compute.googleapis.com'
    ),
    ResourceTestCase(
        resource_data={
            'name': test_resource_name,
            'location': 'us-central1',
            'project_id': test_project
        },
        cls=GcpDatafusionInstance,
        resource_type='datafusion.googleapis.com/Instance',
        name='//datafusion.googleapis.com/projects/my_project/locations/us-central1/instances/my_resource'
    ),
    ResourceTestCase(
        resource_data={
            'name': test_resource_name,
            'location': 'global',
            'project_id': test_project
        },
        cls=GcpDataprocCluster,
        resource_type='dataproc.googleapis.com/Cluster',
        name='//dataproc.googleapis.com/projects/my_project/regions/global/clusters/my_resource'
    ),
    ResourceTestCase(
        resource_data={
            'name': test_resource_name,
            'project_id': test_project
        },
        cls=GcpPubsubSubscription,
        resource_type='pubsub.googleapis.com/Subscription',
        name='//pubsub.googleapis.com/projects/my_project/subscriptions/my_resource'
    ),
    ResourceTestCase(
        resource_data={
            'name': test_resource_name,
            'project_id': test_project
        },
        cls=GcpPubsubTopic,
        resource_type='pubsub.googleapis.com/Topic',
        name='//pubsub.googleapis.com/projects/my_project/topics/my_resource'
    ),
    ResourceTestCase(
        resource_data={
            'name': test_resource_name,
            'project_id': test_project
        },
        cls=GcpSqlInstance,
        resource_type='sqladmin.googleapis.com/Instance',
        name='//cloudsql.googleapis.com/projects/my_project/instances/my_resource'
    ),
    ResourceTestCase(
        resource_data={
            'name': test_resource_name,
            'project_id': test_project
        },
        cls=GcpStorageBucket,
        resource_type='storage.googleapis.com/Bucket',
        # This should include the collection name `/buckets/`, but CAI doesn't do that
        # See: https://issuetracker.google.com/issues/131586763
        name='//storage.googleapis.com/my_resource'
    ),
    ResourceTestCase(
        resource_data={
            'name': test_resource_name,
            'location': 'us-central1',
            'project_id': test_project
        },
        cls=GcpComputeSubnetwork,
        resource_type='compute.googleapis.com/Subnetwork',
        name='//compute.googleapis.com/projects/my_project/regions/us-central1/subnetworks/my_resource'
    ),
    ResourceTestCase(
        resource_data={
            'name': test_resource_name,
            'project_id': test_project
        },
        cls=GcpComputeFirewall,
        resource_type='compute.googleapis.com/Firewall',
        name='//compute.googleapis.com/projects/my_project/global/firewalls/my_resource'
    ),
    ResourceTestCase(
        resource_data={
            'name': test_resource_name,
            'location': 'us-central1',
            'project_id': test_project
        },
        cls=GcpDataflowJob,
        resource_type='dataflow.googleapis.com/Job',
        name='//dataflow.googleapis.com/projects/my_project/locations/us-central1/jobs/my_resource'
    ),
    ResourceTestCase(
        resource_data={
            'name': test_resource_name,
            'location': 'us-central1',
            'project_id': test_project
        },
        cls=GcpRedisInstance,
        resource_type='redis.googleapis.com/Instance',
        name='//redis.googleapis.com/projects/my_project/locations/us-central1/instances/my_resource'
    ),
    ResourceTestCase(
        resource_data={
            'name': test_resource_name,
            'location': 'us-central1',
            'project_id': test_project
        },
        cls=GcpMemcacheInstance,
        resource_type='memcache.googleapis.com/Instance',
        name='//memcache.googleapis.com/projects/my_project/locations/us-central1/instances/my_resource'
    ),
]


@pytest.mark.parametrize(
    "case",
    test_cases,
    ids=[case.cls.__name__ for case in test_cases])
def test_gcp_from_resource(case):
    r = GoogleAPIResource.from_resource_data(resource_type=case.resource_type, client_kwargs=client_kwargs, **case.resource_data)
    assert r.__class__ == case.cls
    assert isinstance(r._get_request_args(), dict)


def test_gcp_from_resource_no_type():
    with pytest.raises(TypeError) as excinfo:
        GoogleAPIResource.from_resource_data(project_id=test_project)

    assert 'required keyword-only argument' in str(excinfo.value)
    assert 'resource_type' in str(excinfo.value)


def test_gcp_resource_bad_type():
    with pytest.raises(ResourceException) as excinfo:
        GoogleAPIResource.from_resource_data(resource_type='fake.type')

    assert 'Unrecognized resource type' in str(excinfo.value)


@pytest.mark.parametrize(
    "case",
    test_cases,
    ids=[case.cls.__name__ for case in test_cases])
def test_gcp_full_resource_name(case):
    r = GoogleAPIResource.from_resource_data(resource_type=case.resource_type, client_kwargs=client_kwargs, **case.resource_data)
    assert r.full_resource_name() == case.name


def test_missing_resource_data():
    with pytest.raises(ResourceException) as excinfo:

        GcpAppEngineInstance(name=test_resource_name)

    assert "Missing data required for resource creation" in str(excinfo.value)


def test_gcp_to_dict():
    r = GoogleAPIResource.from_resource_data(
        resource_type='storage.googleapis.com/Bucket',
        client_kwargs=client_kwargs,
        name=test_resource_name,
        project_id=test_project,
    )

    data = r.to_dict()
    # with no creds, we should still get this key but it should be none
    assert data['project_id'] == test_project



import json
import os

import pytest

from rpe.exceptions import ExtractorException
from rpe.extractors.gcp_auditlogs import GCPAuditLog
from rpe.resources.gcp import GcpComputeDisk


def get_test_data(filename):
    '''Load json data from the tests dir'''
    p = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        'data',
        filename,
    )

    with open(p) as f:
        return json.load(f)


# parameters for testing logs that should return a single asset
test_single_asset_log_params = [
    # filename, expected_resource_type, expected_operation_type, expected_resource_name
    ("app-engine-debug.json", "appengine.googleapis.com/Instance", "update", "aef-default-test-instance"),
    ("bq-ds-set-iam-policy.json", "bigquery.googleapis.com/Dataset", "update", "wooo"),
    ("bigtable-set-iam-policy.json", "bigtableadmin.googleapis.com/Instance", "update", "example-instance"),
    ("pubsub-subscription-set-iam-policy.json", "pubsub.googleapis.com/Subscription", "update", "test-subscription"),
    ("pubsub-topic-set-iam-policy.json", "pubsub.googleapis.com/Topic", "update", "test-topic"),

    # CloudSQL logs are inconsistent. See https://issuetracker.google.com/issues/137629452
    ("cloudsql-resource.labels.json", "sqladmin.googleapis.com/Instance", "create", "test-instance"),
    ("cloudsql-protoPayload.request.body.json", "sqladmin.googleapis.com/Instance", "create", "test-instance"),
    (
        "cloudsql-protoPayload.request.resource.instanceName.instanceId.json", "sqladmin.googleapis.com/Instance",
        "create", "test-instance"
    ),
    ("cloudfunctions-set-iam-policy.json", "cloudfunctions.googleapis.com/CloudFunction", "update", "example_function"),
    ("compute_networks_insert_1.json", "compute.googleapis.com/Network", "create", "xyz"),
    ("compute_networks_insert_2.json", "compute.googleapis.com/Network", "create", "xyzcli"),
    ("compute_networks_insert_3.json", "compute.googleapis.com/Network", "update", "datalab-network"),
    ("compute-subnetworks-enable-flow-logs.json", "compute.googleapis.com/Subnetwork", "update", "example"),
    ("compute-subnetworks-set-private-ip-google-access.json", "compute.googleapis.com/Subnetwork", "update", "example"),
    ("compute-firewalls-enable-logs-policy.json", "compute.googleapis.com/Firewall", "create", "test-firewall"),
    ("dataproc_createcluster.json", "dataproc.googleapis.com/Cluster", "create", "test-dataproc-cluster"),
    ("datafusion-create-instance.json", "datafusion.googleapis.com/Instance", "create", "test-instance"),
    ("datafusion-update-instance.json", "datafusion.googleapis.com/Instance", "update", "test-instance"),
    ("gke-cluster-update.json", "container.googleapis.com/Cluster", "update", "example-cluster"),
    ("gke-nodepool-set.json", "container.googleapis.com/NodePool", "update", "example-pool"),
    ("project_get_iam.json", "cloudresourcemanager.googleapis.com/Project", "read", "cd-np-test1"),
    ("project_set_iam.json", "cloudresourcemanager.googleapis.com/Project", "update", "cd-np-test1"),
    (
        "servicemanagement-enable-service.json", "serviceusage.googleapis.com/Service", "update",
        "youtubeadsreach.googleapis.com"
    ),
    (
        "servicemanagement-disable-service.json", "serviceusage.googleapis.com/Service", "update",
        "youtubereporting.googleapis.com"
    ),
    (
        "servicemanagement-activate-service.json", "serviceusage.googleapis.com/Service", "update",
        "calendar-json.googleapis.com"
    ),
    (
        "servicemanagement-deactivate-service.json", "serviceusage.googleapis.com/Service", "update",
        "zync.googleapis.com"
    ),
    ("serviceusage-enable.json", "serviceusage.googleapis.com/Service", "update", "youtubereporting.googleapis.com"),
    ("serviceusage-disable.json", "serviceusage.googleapis.com/Service", "update", "zync.googleapis.com"),
    ("storage_bucket_delete.json", "storage.googleapis.com/Bucket", "delete", "test-bucket"),
    ("storage_bucket_update.json", "storage.googleapis.com/Bucket", "update", "test-bucket"),
    ("dataflow-job-step.json", "dataflow.googleapis.com/Job", "create", "job-id"),
    ("memorystore-redis.json", "redis.googleapis.com/Instance", "create", "test-instance"),
]

test_log_resource_count_params = [
    ("serviceusage-batchenable.json", 3),
    ("compute-hardened-images.json", 3),
]


@pytest.mark.parametrize(
    "filename,expected_resource_type,expected_operation_type,expected_resource_name", test_single_asset_log_params
)
def test_single_asset_log_messages(filename, expected_resource_type, expected_operation_type, expected_resource_name):
    log_message = get_test_data(filename)

    extracted = GCPAuditLog.extract(log_message)

    assert len(extracted.resources) == 1
    resource = extracted.resources[0]

    assert resource.resource_type == expected_resource_type
    assert extracted.metadata.operation == expected_operation_type
    assert resource.resource_name == expected_resource_name


@pytest.mark.parametrize("filename,expected_resource_count", test_log_resource_count_params)
def test_log_resource_count(filename, expected_resource_count):
    log_message = get_test_data(filename)

    extracted = GCPAuditLog.extract(log_message)
    assert len(extracted.resources) == expected_resource_count


@pytest.mark.parametrize(
    'filename,expected_disk_names', [
        ('compute_instance_creation_logs_1.json', ['console-devicenames', 'disk2name']),
        ('compute_instance_creation_logs_2.json', []),
        ('compute_instance_creation_logs_3.json', ['console-nochange']),
        ('compute_instance_creation_logs_4.json', ['gcloud-devicename-mismatch']),
        ('compute_instance_creation_logs_5.json', ['demo-testing-good-boot', 'demo-testing-good-data']),
    ]
)
def test_compute_instance_logs_get_disk_names(filename, expected_disk_names):
    '''
        Instance creation audit logs contain details about the disks, but extracting the disk name is
        complicated. This tests multiple known log formats for instance creation against expected disk names.
    '''

    log = get_test_data(filename)

    extracted = GCPAuditLog.extract(log)

    disk_names = [
        resource.resource_name for resource in extracted.resources if resource.type() == GcpComputeDisk.resource_type
    ]
    assert sorted(disk_names) == sorted(expected_disk_names)


def test_invalid_audit_log():
    with pytest.raises(ExtractorException):
        extracted = GCPAuditLog.extract({})



# Copyright 2019 The resource-policy-evaluation-library Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import importlib.util
import inspect
import sys

from rpe.policy import Evaluation, Policy


class PythonPolicyEngine:

    counter = 0

    def __init__(self, package_path):

        self._policies = {}
        self.package_path = package_path
        PythonPolicyEngine.counter += 1
        self.package_name = 'rpe.plugins.policies.py_' + str(PythonPolicyEngine.counter)

        self._load_policies()

    def _load_policies(self):
        spec = importlib.util.spec_from_file_location(
            self.package_name,
            "{}/__init__.py".format(self.package_path)
        )
        module = importlib.util.module_from_spec(spec)
        sys.modules[self.package_name] = module

        spec.loader.exec_module(module)

        for name, obj in inspect.getmembers(module):
            if inspect.isclass(obj) and hasattr(obj, 'applies_to') and isinstance(obj.applies_to, list):
                self._policies[name] = obj

    def policies(self):
        """
        Returns:
            A list of names of configured policies
        """

        policies = [
            Policy(
                policy_id=policy_name,
                engine=self,
                applies_to=policy_cls.applies_to,
                description=policy_cls.description
            )
            for policy_name, policy_cls in self._policies.items()
        ]

        return policies

    def evaluate(self, resource):
        matched_policies = dict(filter(
            lambda policy: resource.type() in policy[1].applies_to,
            self._policies.items()
        ))

        # Loop over policy and build evals, so we can catch exceptions

        evals = []

        for policy_name, policy_cls in matched_policies.items():
            try:

                # Ensure that these are boolean
                compliant = policy_cls.compliant(resource) is True
                excluded = policy_cls.excluded(resource) is True

                ev = Evaluation(
                    resource=resource,
                    engine=self,
                    policy_id=policy_name,
                    compliant=compliant,
                    excluded=excluded,
                    remediable=hasattr(policy_cls, 'remediate')
                )

                evals.append(ev)

            # These are user-provided modules, we need to catch any exception
            except Exception as e:
                print(f'Evaluation exception. Policy: {policy_name}, Message: {str(e)}')

        return evals

    def remediate(self, resource, policy_id):
        policy_cls = self._policies[policy_id]
        policy_cls.remediate(resource)



# Copyright 2019 The resource-policy-evaluation-library Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
import tenacity

from urllib import request

from rpe.policy import Evaluation, Policy
from rpe.exceptions import is_retryable_exception
from rpe.exceptions import NoSuchEndpoint
from rpe.exceptions import NoPossibleRemediation

from .base import Engine


class OpenPolicyAgent(Engine):

    def __init__(self, opa_base_url):
        self.opa_base_url = opa_base_url

    @tenacity.retry(
        retry=tenacity.retry_if_exception(is_retryable_exception),
        wait=tenacity.wait_random_exponential(multiplier=1, max=10),
        stop=tenacity.stop_after_attempt(5)
    )
    def _opa_request(self, path, method='GET', data=None):
        url = '{}/{}'.format(self.opa_base_url, path)
        headers = {'Content-type': 'application/json'}
        req = request.Request(
            url,
            data=json.dumps(data).encode('utf-8'),
            method=method,
            headers=headers
        )

        with request.urlopen(req) as resp:
            decoded_resp = resp.read().decode('utf-8')
            deserialized_resp = json.loads(decoded_resp)
            if 'result' not in deserialized_resp:
                err = "Endpoint {} not found on the OPA server.".format(url)
                raise NoSuchEndpoint(err)

            return deserialized_resp['result']

    # Perform an evaluation on a given resource
    def evaluate(self, resource):
        input = {
            'input': resource.get(),
        }

        evals = self._opa_request('rpe/evaluate', method='POST', data=input)

        return [
            Evaluation(engine=self, resource=resource, **ev)
            for ev in evals
        ]

    def policies(self):
        """
        Returns:
            A list of all configured policies, optionally filtered by a resource_type
        """
        policies = self._opa_request('rpe/policies')

        return [
            Policy(engine=self, **p)
            for p in policies
        ]

    def remediate(self, resource, policy_id):
        rem_path = 'rpe/policy/{}/remediate'.format(
            policy_id
        )
        input = {'input': resource.get()}
        remediation = self._opa_request(rem_path, method='POST', data=input)

        if remediation:
            resource.remediate(remediation)
        else:
            raise NoPossibleRemediation("Remediation is not supported for this resource/policy")



# Copyright 2021 The resource-policy-evaluation-library Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Optional

import dateutil.parser
import jmespath

from rpe.exceptions import ExtractorException
from rpe.extractors import Extractor
from rpe.extractors.models import ExtractedMetadata, ExtractedResources
from rpe.resources.gcp import GoogleAPIResource


@dataclass
class OperationEnum(str, Enum):
    create = 'create'
    read = 'read'
    update = 'update'
    delete = 'delete'


@dataclass
class GCPAuditLogMetadata(ExtractedMetadata):
    id: str
    timestamp: datetime
    principal: Optional[str]
    method_name: Optional[str]
    operation: Optional[OperationEnum]


class GCPAuditLog(Extractor):
    @classmethod
    def extract(cls, log_message):
        # Attempt to load it like a pubsub message, without requiring the pubsub client to be installed
        if not isinstance(log_message, dict):
            message_data = json.loads(log_message.data)
        else:
            message_data = log_message

        if not cls.is_audit_log(message_data):
            raise ExtractorException('Not an audit log')

        metadata = cls.get_metadata(message_data)
        resources = cls.get_resources(message_data)

        return ExtractedResources(
            resources=resources,
            metadata=metadata,
        )

    @classmethod
    def is_audit_log(cls, message_data):

        log_type = jmespath.search('protoPayload."@type"', message_data)
        log_name = message_data.get('logName', '')

        # normal activity logs have logName in this form:
        #  projects/<p>/logs/cloudaudit.googleapis.com%2Factivity
        # data access logs have a logName field that looks like:
        #  projects/<p>/logs/cloudaudit.googleapis.com%2Fdata_access

        return all(
            [
                log_type == 'type.googleapis.com/google.cloud.audit.AuditLog',
                log_name.split('/')[-1].startswith('cloudaudit.googleapis.com'),
            ]
        )

    @classmethod
    def get_metadata(cls, message_data):

        method_name = jmespath.search('protoPayload.methodName', message_data)
        insert_id = message_data.get('insertId')

        log_timestamp = message_data.get('timestamp')
        timestamp = dateutil.parser.parse(log_timestamp)

        principal_email = jmespath.search('protoPayload.authenticationInfo.principalEmail', message_data)

        operation = cls.get_operation_type(method_name)

        return GCPAuditLogMetadata(
            id=insert_id,
            timestamp=timestamp,
            method_name=method_name,
            operation=operation,
            principal=principal_email,
        )

    @classmethod
    def get_operation_type(cls, method_name):

        last = method_name.split('.')[-1].lower()
        # For batch methods, look for the verb after the word 'batch'
        if last.startswith('batch'):
            last = last[5:]

        read_prefixes = ('get', 'list')
        if last.startswith(read_prefixes):
            return 'read'

        update_prefixes = (
            'update',
            'patch',
            'set',
            'debug',
            'enable',
            'disable',
            'expand',
            'deactivate',
            'activate',
            'switch',
        )

        if last.startswith(update_prefixes):
            return 'update'

        create_prefixes = (
            'create',
            'insert',
        )

        if last.startswith(create_prefixes):
            return 'create'

        delete_prefixes = ('delete',)
        if last.startswith(delete_prefixes):
            return 'delete'

        return None

    @classmethod
    def get_resources(cls, message):

        resources = []

        res_type = jmespath.search('resource.type', message)
        if res_type is None:
            return resources

        # just shortening the many calls to jmespath throughout this function
        # this sub-function saves us from passing the message each time
        def prop(exp):
            return jmespath.search(exp, message)

        def add_resource():
            r = resource_data.copy()
            res = GoogleAPIResource.from_resource_data(**r)
            resources.append(res)

        method_name = prop('protoPayload.methodName')

        if res_type == 'cloudsql_database' and method_name.startswith('cloudsql.instances'):

            resource_data = {
                'resource_type': 'sqladmin.googleapis.com/Instance',
                # CloudSQL logs are inconsistent. See https://issuetracker.google.com/issues/137629452
                'name': (
                    prop('resource.labels.database_id').split(':')[-1]
                    or prop('protoPayload.request.body.name')
                    or prop('protoPayload.request.resource.instanceName.instanceId')
                ),
                'location': prop('resource.labels.region'),
                'project_id': prop('resource.labels.project_id'),
            }
            add_resource()

        elif res_type == "gcs_bucket" and method_name.startswith(('storage.buckets', 'storage.setIamPermissions')):
            resource_data = {
                'resource_type': 'storage.googleapis.com/Bucket',
                'name': prop("resource.labels.bucket_name"),
                'location': prop("resource.labels.location"),
                'project_id': prop("resource.labels.project_id"),
            }
            add_resource()

        elif res_type == "bigquery_dataset":
            if "DatasetService" in method_name or 'etIamPolicy' in method_name:
                resource_data = {
                    'resource_type': 'bigquery.googleapis.com/Dataset',
                    'name': prop("resource.labels.dataset_id"),
                    'project_id': prop("resource.labels.project_id"),
                }
                add_resource()

        elif res_type == "project" and 'etIamPolicy' in method_name:
            resource_data = {
                'resource_type': 'cloudresourcemanager.googleapis.com/Project',
                'name': prop("resource.labels.project_id"),
                'project_id': prop("resource.labels.project_id"),
            }
            add_resource()

        elif res_type == "pubsub_subscription" and 'etIamPolicy' in method_name:
            resource_data = {
                'resource_type': 'pubsub.googleapis.com/Subscription',
                'name': prop("resource.labels.subscription_id").split('/')[-1],
                'project_id': prop("resource.labels.project_id"),
            }
            add_resource()

        elif res_type == "pubsub_topic" and 'etIamPolicy' in method_name:
            resource_data = {
                'resource_type': 'pubsub.googleapis.com/Topic',
                'name': prop("resource.labels.topic_id").split('/')[-1],
                'project_id': prop("resource.labels.project_id"),
            }
            add_resource()

        elif res_type == 'audited_resource' and (
            'EnableService' in method_name or 'DisableService' in method_name or 'ctivateService' in method_name
        ):

            resource_data = {
                'resource_type': 'serviceusage.googleapis.com/Service',
                'project_id': prop("resource.labels.project_id"),
            }

            # Check if multiple services were included in the request
            # The Google Cloud Console generates (De)activate calls that logs a different format so we check both
            # known formats
            services = prop('protoPayload.request.serviceIds') or prop('protoPayload.request.serviceNames')
            if services:
                for s in services:
                    resource_data['name'] = s
                    add_resource()
            else:
                resource_data['name'] = prop("protoPayload.resourceName").split('/')[-1]
                add_resource()

        elif res_type == 'audited_resource' and 'DeactivateServices' in method_name:
            resource_data = {
                'resource_type': 'serviceusage.googleapis.com/Service',
                'name': prop("resource.labels.service"),
                'project_id': prop("resource.labels.project_id"),
            }
            add_resource()

        elif res_type == "gce_network":
            resource_data = {
                'resource_type': 'compute.googleapis.com/Network',
                'name': prop("protoPayload.resourceName").split('/')[-1],
                'project_id': prop("resource.labels.project_id"),
            }
            add_resource()

        elif res_type == "gce_subnetwork":
            resource_data = {
                'resource_type': 'compute.googleapis.com/Subnetwork',
                'name': prop("resource.labels.subnetwork_name"),
                'project_id': prop("resource.labels.project_id"),
                'location': prop("resource.labels.location"),
            }
            add_resource()

        elif res_type == "gce_firewall_rule":
            resource_data = {
                'resource_type': 'compute.googleapis.com/Firewall',
                'name': prop("protoPayload.resourceName").split('/')[-1],
                'project_id': prop("resource.labels.project_id"),
            }
            add_resource()

        elif res_type == "gae_app" and 'DebugInstance' in method_name:
            instance_data = prop("protoPayload.resourceName").split('/')
            resource_data = {
                'resource_type': 'appengine.googleapis.com/Instance',
                'name': instance_data[-1],
                'app': instance_data[1],
                'service': instance_data[3],
                'version': instance_data[5],
            }
            add_resource()

        elif res_type == "gce_instance":

            instance_name = prop('protoPayload.resourceName').split('/')[-1]

            resource_data = {
                'resource_type': 'compute.googleapis.com/Instance',
                'name': instance_name,
                'location': prop("resource.labels.zone"),
                'project_id': prop("resource.labels.project_id"),
            }

            # Logs are sent for some resources that are hidden by the compute API. We've found that some of these
            # start with reserved prefixes. If the instance looks like a hidden resource, stop looking for
            # resources and return immediately
            compute_reserved_prefixes = ('aef-', 'aet-')
            if resource_data['name'].startswith(compute_reserved_prefixes):
                return resources

            add_resource()

            # Also add disk resources since theres not a separate log message for these
            disks = prop('protoPayload.request.disks') or []

            for disk in disks:

                # The name of the disk is complicated. If the diskName is set in initParams use that
                # If not AND its the boot disk, use the instance name
                # Otherwise use the device name

                disk_name = jmespath.search('initializeParams.diskName', disk)
                device_name = jmespath.search('deviceName', disk)
                boot = jmespath.search('boot', disk)

                actual_disk_name = disk_name or (boot and instance_name) or device_name

                resource_data = {
                    'resource_type': 'compute.googleapis.com/Disk',
                    'name': actual_disk_name,
                    'location': prop("resource.labels.zone"),
                    'project_id': prop("resource.labels.project_id"),
                }

                add_resource()

        elif res_type == "cloud_function":
            resource_data = {
                'name': prop("resource.labels.function_name"),
                'project_id': prop("resource.labels.project_id"),
                'location': prop("resource.labels.region"),
                'resource_type': 'cloudfunctions.googleapis.com/CloudFunction',
            }

            add_resource()

        elif res_type == "cloud_dataproc_cluster":
            resource_data = {
                'resource_type': 'dataproc.googleapis.com/Cluster',
                'project_id': prop("resource.labels.project_id"),
                'name': prop("resource.labels.cluster_name"),
                'location': prop("resource.labels.region"),
            }
            add_resource()

        elif res_type == "gke_cluster":
            resource_data = {
                'resource_type': 'container.googleapis.com/Cluster',
                'name': prop("resource.labels.cluster_name"),
                'project_id': prop("resource.labels.project_id"),
                'location': prop("resource.labels.location"),
            }
            add_resource()

            # add node pool resources for eval on new cluster creation
            if "create" in method_name.lower() and prop("protoPayload.request.cluster.nodePools") is not None:
                resource_data['resource_type'] = 'container.googleapis.com/NodePool'
                resource_data['cluster'] = prop("resource.labels.cluster_name")
                for pool in prop("protoPayload.request.cluster.nodePools"):
                    resource_data['name'] = pool.get('name')
                    add_resource()

        elif res_type == "gke_nodepool":
            resource_data = {
                'resource_type': 'container.googleapis.com/NodePool',
                'cluster': prop("resource.labels.cluster_name"),
                'name': prop("resource.labels.nodepool_name"),
                'project_id': prop("resource.labels.project_id"),
                'location': prop("resource.labels.location"),
            }
            add_resource()

        elif res_type == "audited_resource" and 'BigtableInstanceAdmin' in method_name:
            resource_data = {
                'name': prop("protoPayload.resourceName").split('/')[-1],
                'project_id': prop("resource.labels.project_id"),
            }

            resource_data['resource_type'] = 'bigtableadmin.googleapis.com/Instance'
            add_resource()

        elif res_type == "dataflow_step" and 'create' in method_name:
            # The endpoint expects the job id instead of name
            resource_data = {
                'name': prop("protoPayload.request.job_id"),
                'project_id': prop("resource.labels.project_id"),
                'location': prop("resource.labels.region"),
            }

            resource_data['resource_type'] = 'dataflow.googleapis.com/Job'
            add_resource()

        elif res_type == "audited_resource" and 'CloudRedis' in method_name:
            resource_data = {
                'name': prop("protoPayload.resourceName").split('/')[-1],
                'project_id': prop("resource.labels.project_id"),
                'location': prop("protoPayload.resourceLocation.currentLocations")[0],
            }

            resource_data['resource_type'] = 'redis.googleapis.com/Instance'
            add_resource()

        elif res_type == "audited_resource" and prop('resource.labels.service') == 'datafusion.googleapis.com':
            name_bits = prop('protoPayload.resourceName').split('/')
            resource_data = {
                'name': name_bits[5],
                'project_id': name_bits[1],
                'location': name_bits[3],
                'resource_type': 'datafusion.googleapis.com/Instance',
            }
            add_resource()

        return resources
